@@ -18,8 +18,6 @@ class Command(BaseCommand):
         # using kwargs in Python 3.6. But we need Python 3.5 compatibility for
         # now.
         # See https://www.python.org/dev/peps/pep-0468/.
-        # ActivitySynchronizer is omitted here because it typically takes
-        # a long time and activities aren't a common use.
         synchronizers = (
             ('member', sync.MemberSynchronizer, _('Member')),
             ('board', sync.BoardSynchronizer, _('Board')),
@@ -28,7 +26,11 @@ class Command(BaseCommand):
             ('priority', sync.PrioritySynchronizer, _('Priority')),
             ('project_status', sync.ProjectStatusSynchronizer,
                 _('Project Status')),
+            ('project_type', sync.ProjectTypeSynchronizer,
+             _('Project Type')),
             ('project', sync.ProjectSynchronizer, _('Project')),
+            ('project_phase', sync.ProjectPhaseSynchronizer,
+             _('Project Phase')),
             ('territory', sync.TerritorySynchronizer,
                 _('Territory')),
             ('company_status', sync.CompanyStatusSynchronizer,
@@ -56,7 +58,7 @@ class Command(BaseCommand):
                 _('Company Other')),
             ('sla', sync.SLASynchronizer,
                 _('Sla')),
-            ('sla_priority', sync.SLAPrioritySychronizer,
+            ('sla_priority', sync.SLAPrioritySynchronizer,
                 _('Sla Priority')),
             ('type', sync.TypeSynchronizer,
              _('Type')),
@@ -64,19 +66,33 @@ class Command(BaseCommand):
              _('Sub Type')),
             ('item', sync.ItemSynchronizer,
              _('Item')),
-            ('ticket', sync.TicketSynchronizer, _('Ticket')),
+            ('ticket', sync.ServiceTicketSynchronizer, _('Ticket')),
+            ('project_ticket', sync.ProjectTicketSynchronizer,
+                _('Project Ticket')),
             ('service_note', sync.ServiceNoteSynchronizer,
                 _('Service Note')),
             ('opportunity_note', sync.OpportunityNoteSynchronizer,
                 _('Opportunity Note')),
-            ('schedule_type', sync.ScheduleTypeSychronizer,
+            ('work_type', sync.WorkTypeSynchronizer,
+             _('Work Type')),
+            ('work_role', sync.WorkRoleSynchronizer,
+             _('Work Role')),
+            ('agreement', sync.AgreementSynchronizer,
+             _('Agreement')),
+            ('activity_status', sync.ActivityStatusSynchronizer,
+                _('Activity Status')),
+            ('activity_type', sync.ActivityTypeSynchronizer,
+                _('Activity Type')),
+            ('activity', sync.ActivitySynchronizer,
+                _('Activity')),
+            ('schedule_type', sync.ScheduleTypeSynchronizer,
                 _('Schedule Type')),
             ('schedule_status', sync.ScheduleStatusSynchronizer,
                 _('Schedule Status')),
             ('schedule_entry', sync.ScheduleEntriesSynchronizer,
                 _('Schedule Entry')),
             ('time_entry', sync.TimeEntrySynchronizer,
-                _('Time Entry'))
+                _('Time Entry')),
         )
         self.synchronizer_map = OrderedDict()
         for name, synchronizer, obj_name in synchronizers:
@@ -127,14 +143,6 @@ class Command(BaseCommand):
         failed_classes = 0
         error_messages = ''
 
-        num_synchronizers = len(self.synchronizer_map)
-        has_ticket_sync = 'ticket' in self.synchronizer_map
-        if full_option and num_synchronizers and has_ticket_sync:
-            sync_classes = list(sync_classes)
-            sync_classes.reverse()
-            # need to move ticket synchronizer to the tail of the list
-            sync_classes.append(sync_classes.pop(0))
-
         for sync_class, obj_name in sync_classes:
             try:
                 self.sync_by_class(sync_class, obj_name,
