@@ -2,11 +2,14 @@ from copy import deepcopy
 from unittest import TestCase
 from django.test import TransactionTestCase
 from django.core.files.storage import default_storage
+from django.test import override_settings
+
+from botocore.exceptions import NoCredentialsError
 
 import datetime
 
 from dateutil.parser import parse
-from djconnectwise.models import Activity
+from djconnectwise.models import ActivityStatus, ActivityType, Activity
 from djconnectwise.models import BoardStatus
 from djconnectwise.models import Company, CompanyStatus, CompanyType
 from djconnectwise.models import ConnectWiseBoard
@@ -18,7 +21,7 @@ from djconnectwise.models import Opportunity
 from djconnectwise.models import OpportunityStatus
 from djconnectwise.models import OpportunityStage
 from djconnectwise.models import OpportunityType
-from djconnectwise.models import Project, ProjectStatus
+from djconnectwise.models import Project, ProjectStatus, ProjectPhase
 from djconnectwise.models import ScheduleEntry
 from djconnectwise.models import ScheduleStatus
 from djconnectwise.models import ScheduleType
@@ -37,6 +40,10 @@ from djconnectwise.models import MyCompanyOther
 from djconnectwise.models import Type
 from djconnectwise.models import SubType
 from djconnectwise.models import Item
+from djconnectwise.models import WorkType
+from djconnectwise.models import WorkRole
+from djconnectwise.models import Agreement
+from djconnectwise.models import ProjectType
 from djconnectwise.utils import get_hash
 
 from . import fixtures
@@ -94,8 +101,8 @@ class SynchronizerTestMixin:
         self._sync(new_json_list)
 
         changed = self.model_class.objects.get(id=instance_id)
-        self.assertNotEqual(original.name,
-                            name)
+
+        self.assertNotEqual(original.name, name)
         self._assert_fields(changed, new_json)
 
 
@@ -221,6 +228,8 @@ class TestTimeEntrySynchronizer(TestCase, SynchronizerTestMixin):
 
     def setUp(self):
         super().setUp()
+        fixture_utils.init_board_statuses()
+        fixture_utils.init_tickets()
         fixture_utils.init_time_entries()
 
     def call_api(self, return_data):
@@ -340,9 +349,9 @@ class TestScheduleEntriesSynchronizer(TestCase, SynchronizerTestMixin):
         self.assertEqual(instance.id, json_data['id'])
         self.assertEqual(instance.name, json_data['name'])
         self.assertEqual(instance.done_flag, json_data['doneFlag'])
-        self.assertEqual(instance.expected_date_start,
+        self.assertEqual(instance.date_start,
                          parse(json_data['dateStart']))
-        self.assertEqual(instance.expected_date_end,
+        self.assertEqual(instance.date_end,
                          parse(json_data['dateEnd']))
 
         # verify referenced objects
@@ -358,7 +367,7 @@ class TestScheduleEntriesSynchronizer(TestCase, SynchronizerTestMixin):
 
 
 class TestScheduleTypeSynchronizer(TestCase, SynchronizerTestMixin):
-    synchronizer_class = sync.ScheduleTypeSychronizer
+    synchronizer_class = sync.ScheduleTypeSynchronizer
     model_class = ScheduleType
     fixture = fixtures.API_SCHEDULE_TYPE_LIST
 
@@ -399,6 +408,78 @@ class TestProjectStatusSynchronizer(TestCase, SynchronizerTestMixin):
         self.assertEqual(instance.closed_flag, json_data['closedFlag'])
 
 
+class TestProjectTypeSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.ProjectTypeSynchronizer
+    model_class = ProjectType
+    fixture = fixtures.API_PROJECT_TYPES
+
+    def call_api(self, return_data):
+        return mocks.projects_api_get_project_types_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.default_flag, json_data['defaultFlag'])
+        self.assertEqual(instance.inactive_flag, json_data['inactiveFlag'])
+
+
+class TestProjectPhaseSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.ProjectPhaseSynchronizer
+    model_class = ProjectPhase
+    fixture = fixtures.API_PROJECT_PHASE_LIST
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_projects()
+        fixture_utils.init_project_phases()
+
+    def call_api(self, return_data):
+        return mocks.projects_api_get_project_phases_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.description, json_data['description'])
+        self.assertEqual(instance.bill_time, json_data['billTime'])
+        self.assertEqual(instance.notes, json_data['notes'])
+        self.assertEqual(instance.scheduled_hours, json_data['scheduledHours'])
+        self.assertEqual(instance.actual_hours, json_data['actualHours'])
+        self.assertEqual(instance.budget_hours, json_data['budgetHours'])
+        self.assertEqual(instance.project_id, json_data['projectId'])
+
+        self.assertEqual(
+            instance.scheduled_start, parse(json_data['scheduledStart']).date()
+        )
+        self.assertEqual(
+            instance.scheduled_end, parse(json_data['scheduledEnd']).date()
+        )
+        self.assertEqual(
+            instance.actual_start, parse(json_data['actualStart']).date()
+        )
+        self.assertEqual(
+            instance.actual_end, parse(json_data['actualEnd']).date()
+        )
+
+    def test_sync_update(self):
+        self._sync(self.fixture)
+
+        json_data = self.fixture[0]
+
+        instance_id = json_data['id']
+        original = self.model_class.objects.get(id=instance_id)
+
+        description = 'Some New Description'
+        new_json = deepcopy(self.fixture[0])
+        new_json['description'] = description
+        new_json_list = [new_json]
+
+        self._sync(new_json_list)
+
+        changed = self.model_class.objects.get(id=instance_id)
+
+        self.assertNotEqual(original.description, description)
+        self._assert_fields(changed, new_json)
+
+
 class TestProjectSynchronizer(TestCase, SynchronizerTestMixin):
     synchronizer_class = sync.ProjectSynchronizer
     model_class = Project
@@ -428,6 +509,18 @@ class TestProjectSynchronizer(TestCase, SynchronizerTestMixin):
             float(instance.scheduled_hours),
             json_data['scheduledHours']
         )
+        self.assertEqual(
+            instance.actual_start, parse(json_data['actualStart']).date()
+        )
+        self.assertEqual(
+            instance.actual_end, parse(json_data['actualEnd']).date()
+        )
+        self.assertEqual(
+            instance.estimated_start, parse(json_data['estimatedStart']).date()
+        )
+        self.assertEqual(
+            instance.estimated_end, parse(json_data['estimatedEnd']).date()
+        )
 
 
 class TestTeamSynchronizer(TestCase, SynchronizerTestMixin):
@@ -504,6 +597,16 @@ class TestBoardSynchronizer(TestCase, SynchronizerTestMixin):
     def _assert_fields(self, instance, json_data):
         self.assertEqual(instance.name, json_data['name'])
         self.assertEqual(instance.inactive, json_data['inactiveFlag'])
+        self.assertEqual(instance.work_role.name,
+                         json_data['workRole']['name'])
+        self.assertEqual(instance.work_type.name,
+                         json_data['workType']['name'])
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_work_roles()
+        fixture_utils.init_work_types()
+        fixture_utils.init_boards()
 
 
 class TestBoardStatusSynchronizer(TestCase, SynchronizerTestMixin):
@@ -688,6 +791,29 @@ class TestMemberSynchronization(TransactionTestCase):
         self.assertTrue(default_storage.exists(avatar_size))
         self.assertTrue(default_storage.exists(micro_avatar_size))
 
+    @override_settings(DEFAULT_FILE_STORAGE='storages.backends.'
+                                            's3boto3.S3Boto3Storage',
+                       AWS_STORAGE_BUCKET_NAME='somebucket')
+    def test_app_wont_crash_without_DO_credentials(self):
+        self.synchronizer = sync.MemberSynchronizer()
+        api_member = fixtures.API_MEMBER
+
+        # If this doesn't raise an exception, GREAT!
+        self.synchronizer.update_or_create_instance(api_member)
+
+        # Now we check to make sure that it WOULD have raised an exception
+        # if we didn't call it from its safe location. If no exceptions
+        # occur then there is still a problem.
+        self.synchronizer = sync.MemberSynchronizer()
+        self.synchronizer.sync()
+        member = Member.objects.get(identifier=self.identifier)
+
+        attachment_filename = 'some_new_image.png'
+        avatar = mocks.get_member_avatar()
+        with self.assertRaises(NoCredentialsError):
+            self.synchronizer._save_avatar(
+                member, avatar, attachment_filename)
+
 
 class TestOpportunitySynchronizer(TestCase, SynchronizerTestMixin):
     synchronizer_class = sync.OpportunitySynchronizer
@@ -1066,8 +1192,8 @@ class TestSLASynchronizer(TestCase, SynchronizerTestMixin):
                          json_data['resolutionHours'])
 
 
-class TestSLAPrioritySychronizer(TestCase, SynchronizerTestMixin):
-    synchronizer_class = sync.SLAPrioritySychronizer
+class TestSLAPrioritySynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.SLAPrioritySynchronizer
     model_class = SlaPriority
     fixture = fixtures.API_SERVICE_SLA_PRIORITY_LIST
 
@@ -1107,16 +1233,12 @@ class TestSLAPrioritySychronizer(TestCase, SynchronizerTestMixin):
                          json_data['resolutionHours'])
 
 
-class TestTicketSynchronizer(TestCase):
-
+class TestTicketSynchronizerMixin:
     def setUp(self):
         super().setUp()
         mocks.system_api_get_members_call(fixtures.API_MEMBER_LIST)
         mocks.system_api_get_member_image_by_photo_id_call(
             (mocks.CW_MEMBER_IMAGE_FILENAME, mocks.get_member_avatar()))
-        mocks.service_api_tickets_call()
-
-        self._init_data()
 
     def _clean(self):
         Ticket.objects.all().delete()
@@ -1137,6 +1259,196 @@ class TestTicketSynchronizer(TestCase):
         fixture_utils.init_types()
         fixture_utils.init_subtypes()
         fixture_utils.init_items()
+        fixture_utils.init_agreements()
+
+    def test_sync_ticket(self):
+        """
+        Test to ensure ticket synchronizer saves a CW Ticket instance
+        locally.
+        """
+        synchronizer = self.sync_class()
+        synchronizer.sync()
+        self.assertGreater(Ticket.objects.all().count(), 0)
+
+        json_data = self.ticket_fixture
+        instance = Ticket.objects.get(id=json_data['id'])
+
+        self._assert_sync(instance, json_data)
+        assert_sync_job(Ticket)
+
+    def test_sync_ticket_truncates_automatic_cc_field(self):
+        """
+        Test to ensure ticket synchronizer truncates the automatic CC field
+        to 1000 characters.
+        """
+        synchronizer = self.sync_class()
+
+        instance = Ticket()
+
+        field_data = "kanban@kanban.com;"
+        for i in range(6):
+            # Make some field data with length 1152
+            field_data = field_data + field_data
+
+        json_data = deepcopy(self.ticket_fixture)
+        json_data['automaticEmailCc'] = field_data
+        instance = synchronizer._assign_field_data(instance, json_data)
+        self.assertEqual(len(instance.automatic_email_cc), 1000)
+
+    def test_sync_child_tickets(self):
+        """
+        Test to ensure that a ticket will sync related objects,
+        in its case schedule, note, and time entries
+        """
+        fixture_utils.init_schedule_entries()
+        fixture_utils.init_service_notes()
+        synchronizer = self.sync_class()
+
+        ticket = Ticket.objects.get(id=self.ticket_fixture['id'])
+
+        # Change some fields on all child objects
+        updated_fixture = deepcopy(fixtures.API_SERVICE_NOTE_LIST[0])
+        updated_fixture['ticketId'] = ticket.id
+        updated_fixture['text'] = 'Some new text'
+        fixture_list = [updated_fixture]
+
+        method_name = 'djconnectwise.api.ServiceAPIClient.get_notes'
+        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
+
+        updated_fixture = deepcopy(fixtures.API_TIME_ENTRY)
+        updated_fixture['chargeToId'] = ticket.id
+        updated_fixture['text'] = 'Some new text'
+        updated_fixture['timeEnd'] = '2005-05-16T15:00:00Z'
+        fixture_list = [updated_fixture]
+
+        method_name = 'djconnectwise.api.TimeAPIClient.get_time_entries'
+        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
+
+        updated_fixture = deepcopy(fixtures.API_SALES_ACTIVITY)
+        fixture_list = [updated_fixture]
+
+        method_name = 'djconnectwise.api.SalesAPIClient.get_activities'
+        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
+
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_ticket'
+        mock_call, _patch = \
+            mocks.create_mock_call(method_name, self.ticket_fixture)
+
+        # Trigger method called on callback
+        synchronizer.fetch_sync_by_id(ticket.id)
+
+        # Get the new Values from the db
+        updated_note = ServiceNote.objects.filter(ticket=ticket)[0]
+
+        updated_time = TimeEntry.objects.filter(charge_to_id=ticket)[0]
+
+        # Confirm that they have all been updated
+        self.assertEqual('Some new text', updated_note.text)
+        self.assertEqual(
+            datetime.datetime(
+                2005, 5, 16, 15, 0, tzinfo=datetime.timezone.utc),
+            updated_time.time_end
+            )
+
+    def test_sync_updated(self):
+
+        updated_ticket_fixture = deepcopy(self.ticket_fixture)
+        updated_ticket_fixture['summary'] = 'A new kind of summary'
+        fixture_list = [updated_ticket_fixture]
+
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_tickets'
+        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
+        synchronizer = self.sync_class()
+        # Synchronizer is called twice, as we are testing that synchronizers
+        # can be called twice and keep the same behaviour
+        synchronizer.sync()
+        created_count, updated_count, _ = synchronizer.sync()
+
+        self.assertEqual(created_count, 0)
+        self.assertEqual(updated_count, len(fixture_list))
+
+        instance = Ticket.objects.get(id=updated_ticket_fixture['id'])
+        self._assert_sync(instance, updated_ticket_fixture)
+
+    def test_sync_multiple_status_batches(self):
+        sync.MAX_URL_LENGTH = 330
+        sync.MIN_URL_LENGTH = 320
+        self._init_data()
+        fixture_utils.init_tickets()
+        updated_ticket_fixture = deepcopy(self.ticket_fixture)
+        updated_ticket_fixture['summary'] = 'A new kind of summary'
+        fixture_list = [updated_ticket_fixture]
+
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_tickets'
+        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
+        synchronizer = sync.ServiceTicketSynchronizer()
+        synchronizer.batch_condition_list.extend(
+            [234234, 345345, 234213, 2344523, 345645]
+        )
+        created_count, updated_count, _ = synchronizer.sync()
+
+        self.assertEqual(mock_call.call_count, 2)
+
+    def test_delete_stale_tickets(self):
+        """Local ticket should be deleted if omitted from sync"""
+        ticket_id = self.ticket_fixture['id']
+        ticket_qset = Ticket.objects.filter(id=ticket_id)
+        self.assertEqual(ticket_qset.count(), 1)
+
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_tickets'
+        mock_call, _patch = mocks.create_mock_call(method_name, [])
+        synchronizer = self.sync_class(full=True)
+        synchronizer.sync()
+        self.assertEqual(ticket_qset.count(), 0)
+        _patch.stop()
+
+    def test_callback_sync(self):
+        # Sync initial time entry
+        mocks.time_api_get_time_entries_call(fixtures.API_TIME_ENTRY_LIST)
+        time_entry_sync = sync.TimeEntrySynchronizer()
+        time_entry_sync.sync()
+        self.assertGreater(
+            SyncJob.objects.filter(entity_name='TimeEntry').count(), 0
+        )
+
+        # Mock the child class syncs
+        mocks.service_api_get_notes_call(fixtures.API_SERVICE_NOTE_LIST)
+        mocks.sales_api_get_activities_call(fixtures.API_SALES_ACTIVITIES)
+
+        method_name = 'djconnectwise.sync.TicketSynchronizerMixin.get_single'
+        mock_call, _patch = \
+            mocks.create_mock_call(method_name, self.ticket_fixture)
+
+        # Create new time entry to sync
+        new_time_entry = deepcopy(fixtures.API_TIME_ENTRY)
+        new_time_entry['id'] = 3
+        mocks.time_api_get_time_entries_call([new_time_entry,
+                                              fixtures.API_TIME_ENTRY])
+
+        ticket_id = self.ticket_fixture['id']
+        synchronizer = self.sync_class()
+        # Simulate ticket getting updated by a callback
+        synchronizer.fetch_sync_by_id(ticket_id)
+
+        # Verify that no time entries are removed,
+        # and that only one entry is added
+        last_sync_job = SyncJob.objects.filter(entity_name='TimeEntry').last()
+        self.assertEqual(last_sync_job.deleted, 0)
+        self.assertEqual(last_sync_job.updated, 0)
+        self.assertEqual(last_sync_job.added, 1)
+        self.assertEqual(last_sync_job.sync_type, 'partial')
+
+
+class TestServiceTicketSynchronizer(TestTicketSynchronizerMixin, TestCase):
+    sync_class = sync.ServiceTicketSynchronizer
+    ticket_fixture = fixtures.API_SERVICE_TICKET
+
+    def setUp(self):
+        super().setUp()
+        mocks.service_api_tickets_call()
+
+        self._init_data()
+        fixture_utils.init_tickets()
 
     def _assert_sync(self, instance, json_data):
         self.assertEqual(instance.summary, json_data['summary'])
@@ -1175,10 +1487,6 @@ class TestTicketSynchronizer(TestCase):
         self.assertEqual(instance.location_id,
                          json_data['serviceLocation']['id'])
 
-        # verify assigned project
-        self.assertEqual(instance.project_id,
-                         json_data['project']['id'])
-
         # verify assigned status
         self.assertEqual(instance.status_id,
                          json_data['status']['id'])
@@ -1192,135 +1500,174 @@ class TestTicketSynchronizer(TestCase):
         # verify assigned type
         self.assertEqual(instance.sub_type_item_id, json_data['item']['id'])
 
-    def test_sync_ticket(self):
+        self.assertEqual(instance.bill_time, json_data['billTime'])
+        self.assertEqual(instance.automatic_email_cc_flag,
+                         json_data['automaticEmailCcFlag'])
+        self.assertEqual(instance.automatic_email_contact_flag,
+                         json_data['automaticEmailContactFlag'])
+        self.assertEqual(instance.automatic_email_resource_flag,
+                         json_data['automaticEmailResourceFlag'])
+        self.assertEqual(instance.automatic_email_cc,
+                         json_data['automaticEmailCc'])
+        self.assertEqual(instance.agreement, json_data['agreement'])
+
+    def test_project_tickets_not_deleted_during_sync(self):
         """
-        Test to ensure ticket synchronizer saves a CW Ticket instance
-        locally.
+        Verify that during a sync of service tickets, no project tickets are
+        removed.
         """
-        synchronizer = sync.TicketSynchronizer()
+        synchronizer = self.sync_class(full=True)
         synchronizer.sync()
-        self.assertGreater(Ticket.objects.all().count(), 0)
 
-        json_data = fixtures.API_SERVICE_TICKET
-        instance = Ticket.objects.get(id=json_data['id'])
-        self._assert_sync(instance, json_data)
-        assert_sync_job(Ticket)
+        self.assertTrue(Ticket.objects.get(id=self.ticket_fixture['id']))
 
-    def test_sync_child_tickets(self):
-        """
-        Test to ensure that a ticket will sync related objects,
-        in its case schedule, note, and time entries
-        """
-        self._init_data()
-        fixture_utils.init_tickets()
-        fixture_utils.init_schedule_entries()
-        fixture_utils.init_time_entries()
-        fixture_utils.init_service_notes()
-        synchronizer = sync.TicketSynchronizer()
-
-        ticket = Ticket.objects.get(id=fixtures.API_SERVICE_TICKET['id'])
-
-        # Change some fields on all three child objects
-        updated_fixture = deepcopy(
-            fixtures.API_SCHEDULE_ENTRY_FOR_TICKET
-            )
-        updated_fixture['name'] = 'A new kind of name'
-        fixture_list = [updated_fixture]
-        fixture_list
-
-        method_name = 'djconnectwise.api.ScheduleAPIClient' \
-            '.get_schedule_entries'
-        mock_call, _patch = mocks.create_mock_call(
-            method_name, fixture_list
+        project_ticket = Ticket.objects.create(
+            summary='Project ticket',
+            record_type='ProjectTicket'
         )
+        project_ticket.save()
 
-        updated_fixture = deepcopy(fixtures.API_SERVICE_NOTE_LIST[0])
-        updated_fixture['text'] = 'Some new text'
-        fixture_list = [updated_fixture]
-
-        method_name = 'djconnectwise.api.ServiceAPIClient.get_notes'
-        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
-
-        updated_fixture = deepcopy(fixtures.API_TIME_ENTRY)
-        updated_fixture['timeEnd'] = '2005-05-16T15:00:00Z'
-        fixture_list = [updated_fixture]
-
-        method_name = 'djconnectwise.api.TimeAPIClient.get_time_entries'
-        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
-
-        # Trigger method called on callback
-        synchronizer.fetch_sync_by_id(ticket.id)
-
-        # Get the new Values from the db
-        updated_sched = ScheduleEntry.objects.filter(ticket_object=ticket)[0]
-
-        updated_note = ServiceNote.objects.filter(ticket=ticket)[0]
-
-        updated_time = TimeEntry.objects.filter(charge_to_id=ticket)[0]
-
-        # Confirm that they have all been updated
-        self.assertEqual('A new kind of name', updated_sched.name)
-        self.assertEqual('Some new text', updated_note.text)
-        self.assertEqual(
-            datetime.datetime(
-                2005, 5, 16, 15, 0, tzinfo=datetime.timezone.utc),
-            updated_time.time_end
-            )
-
-    def test_sync_updated(self):
-        self._init_data()
-        fixture_utils.init_tickets()
-        updated_ticket_fixture = deepcopy(fixtures.API_SERVICE_TICKET)
-        updated_ticket_fixture['summary'] = 'A new kind of summary'
-        fixture_list = [updated_ticket_fixture]
-
-        method_name = 'djconnectwise.api.ServiceAPIClient.get_tickets'
-        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
-        synchronizer = sync.TicketSynchronizer()
-        # Synchronizer is called twice, as we are testing that synchronizers
-        # can be called twice and keep the same behaviour
-        synchronizer.sync()
-        created_count, updated_count, _ = synchronizer.sync()
-
-        self.assertEqual(created_count, 0)
-        self.assertEqual(updated_count, len(fixture_list))
-
-        instance = Ticket.objects.get(id=updated_ticket_fixture['id'])
-        self._assert_sync(instance, updated_ticket_fixture)
-
-    def test_sync_multiple_status_batches(self):
-        sync.MAX_URL_LENGTH = 330
-        sync.MIN_URL_LENGTH = 320
-        self._init_data()
-        fixture_utils.init_tickets()
-        updated_ticket_fixture = deepcopy(fixtures.API_SERVICE_TICKET)
-        updated_ticket_fixture['summary'] = 'A new kind of summary'
-        fixture_list = [updated_ticket_fixture]
-
-        method_name = 'djconnectwise.api.ServiceAPIClient.get_tickets'
-        mock_call, _patch = mocks.create_mock_call(method_name, fixture_list)
-        synchronizer = sync.TicketSynchronizer()
-        synchronizer.batch_condition_list.extend(
-            [234234, 345345, 234213, 2344523, 345645]
-        )
-        created_count, updated_count, _ = synchronizer.sync()
-
-        self.assertEqual(mock_call.call_count, 2)
-
-    def test_delete_stale_tickets(self):
-        """Local ticket should be deleted if omitted from sync"""
-        fixture_utils.init_tickets()
-
-        ticket_id = fixtures.API_SERVICE_TICKET['id']
-        ticket_qset = Ticket.objects.filter(id=ticket_id)
-        self.assertEqual(ticket_qset.count(), 1)
-
-        method_name = 'djconnectwise.api.ServiceAPIClient.get_tickets'
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_tickets'
         mock_call, _patch = mocks.create_mock_call(method_name, [])
-        synchronizer = sync.TicketSynchronizer(full=True)
+        synchronizer = self.sync_class(full=True)
         synchronizer.sync()
-        self.assertEqual(ticket_qset.count(), 0)
-        _patch.stop()
+
+        # Verify that the service ticket has been removed and project ticket
+        # still exists.
+        self.assertEqual(
+            Ticket.objects.get(id=project_ticket.id), project_ticket)
+        self.assertFalse(
+            Ticket.objects.filter(id=self.ticket_fixture['id']).exists())
+
+
+class TestProjectTicketSynchronizer(TestTicketSynchronizerMixin, TestCase):
+    sync_class = sync.ProjectTicketSynchronizer
+    ticket_fixture = fixtures.API_PROJECT_TICKET
+
+    def setUp(self):
+        super().setUp()
+        mocks.project_api_tickets_call()
+
+        self._init_data()
+        fixture_utils.init_project_tickets()
+
+    def _assert_sync(self, instance, json_data):
+        self.assertEqual(instance.summary, json_data['summary'])
+        self.assertEqual(instance.closed_flag, json_data.get('closedFlag'))
+        self.assertEqual(instance.last_updated_utc,
+                         parse(json_data.get('_info').get('lastUpdated')))
+        self.assertEqual(instance.required_date_utc,
+                         parse(json_data.get('requiredDate')))
+        self.assertEqual(instance.resources, json_data.get('resources'))
+        self.assertEqual(instance.budget_hours, json_data.get('budgetHours'))
+        self.assertEqual(instance.actual_hours, json_data.get('actualHours'))
+
+        # verify assigned board
+        self.assertEqual(instance.board_id, json_data['board']['id'])
+
+        # verify assigned company
+        self.assertEqual(instance.company_id, json_data['company']['id'])
+
+        # verify assigned priority
+        self.assertEqual(instance.priority_id, json_data['priority']['id'])
+
+        # verify assigned location
+        self.assertEqual(instance.location_id,
+                         json_data['serviceLocation']['id'])
+
+        # verify assigned project
+        self.assertEqual(instance.project_id,
+                         json_data['project']['id'])
+
+        # verify assigned status
+        self.assertEqual(instance.status_id,
+                         json_data['status']['id'])
+
+        # verify assigned type
+        self.assertEqual(instance.type_id, json_data['type']['id'])
+
+        # verify assigned subtype
+        self.assertEqual(instance.sub_type_id, json_data['subType']['id'])
+
+        # verify assigned subtype item
+        self.assertEqual(instance.sub_type_item_id, json_data['item']['id'])
+
+        self.assertEqual(instance.bill_time, json_data['billTime'])
+        self.assertEqual(instance.automatic_email_cc_flag,
+                         json_data['automaticEmailCcFlag'])
+        self.assertEqual(instance.automatic_email_contact_flag,
+                         json_data['automaticEmailContactFlag'])
+        self.assertEqual(instance.automatic_email_resource_flag,
+                         json_data['automaticEmailResourceFlag'])
+        self.assertEqual(instance.automatic_email_cc,
+                         json_data['automaticEmailCc'])
+        self.assertEqual(instance.agreement, json_data['agreement'])
+
+    def test_service_tickets_not_deleted_during_sync(self):
+        """
+        Verify that during a sync of project tickets, no service tickets are
+        removed.
+        """
+        synchronizer = self.sync_class(full=True)
+        synchronizer.sync()
+        self.assertTrue(Ticket.objects.get(id=self.ticket_fixture['id']))
+
+        service_ticket = Ticket.objects.create(
+            summary='Service ticket',
+            record_type='ServiceTicket'
+        )
+        service_ticket.save()
+
+        method_name = 'djconnectwise.api.TicketAPIMixin.get_tickets'
+        mock_call, _patch = mocks.create_mock_call(method_name, [])
+        synchronizer = self.sync_class(full=True)
+        synchronizer.sync()
+
+        # Verify that the project ticket has been removed and service ticket
+        # still exists.
+        self.assertEqual(
+            Ticket.objects.get(id=service_ticket.id), service_ticket)
+        self.assertFalse(
+            Ticket.objects.filter(id=self.ticket_fixture['id']).exists())
+
+
+class TestActivityStatusSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.ActivityStatusSynchronizer
+    model_class = ActivityStatus
+    fixture = fixtures.API_SALES_ACTIVITY_STATUSES
+
+    def call_api(self, return_data):
+        return mocks.sales_api_get_activities_statuses_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.default_flag, json_data['defaultFlag'])
+        self.assertEqual(instance.inactive_flag, json_data['inactiveFlag'])
+        self.assertEqual(
+            instance.spawn_followup_flag,
+            json_data.get('spawnFollowupFlag', False)
+        )
+        self.assertEqual(instance.closed_flag, json_data['closedFlag'])
+
+
+class TestActivityTypeSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.ActivityTypeSynchronizer
+    model_class = ActivityType
+    fixture = fixtures.API_SALES_ACTIVITY_TYPES
+
+    def call_api(self, return_data):
+        return mocks.sales_api_get_activities_types_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.default_flag, json_data['defaultFlag'])
+        self.assertEqual(instance.inactive_flag, json_data['inactiveFlag'])
+        self.assertEqual(instance.email_flag, json_data['emailFlag'])
+        self.assertEqual(instance.memo_flag, json_data['memoFlag'])
+        self.assertEqual(instance.history_flag, json_data['historyFlag'])
 
 
 class TestActivitySynchronizer(TestCase, SynchronizerTestMixin):
@@ -1336,10 +1683,10 @@ class TestActivitySynchronizer(TestCase, SynchronizerTestMixin):
         fixture_utils.init_opportunity_types()
         fixture_utils.init_opportunity_stages()
         fixture_utils.init_opportunities()
+        fixture_utils.init_activity_statuses()
+        fixture_utils.init_activity_types()
+        fixture_utils.init_agreements()
         fixture_utils.init_activities()
-        mocks.sales_api_get_activities_call(
-            fixtures.API_SALES_ACTIVITIES)
-        sync.ActivitySynchronizer().sync()
 
     def call_api(self, return_data):
         return mocks.sales_api_get_activities_call(return_data)
@@ -1365,30 +1712,39 @@ class TestActivitySynchronizer(TestCase, SynchronizerTestMixin):
                          api_activity['opportunity']['id'])
         if api_activity['ticket'] is not None:
             self.assertEqual(activity.ticket_id, api_activity['ticket']['id'])
+        self.assertEqual(
+            activity.status_id, api_activity['status']['id']
+        )
+        self.assertEqual(
+            activity.type_id, api_activity['type']['id']
+        )
+        self.assertEqual(
+            activity.company_id, api_activity['company']['id']
+        )
+        self.assertEqual(
+            activity.agreement_id, api_activity['agreement']['id']
+        )
 
-    # TODO Django 2.0.5 broke this test, but seeing as we aren't syncing
-    #      activities yet, this test may not even be the same at all in the
-    #      future. Commenting out until activity syncing is back in.
-    # def test_sync_null_member_activity(self):
-    #     null_member_activity = deepcopy(fixtures.API_SALES_ACTIVITY)
-    #     null_member_activity['id'] = 999
-    #     null_member_activity['assignTo'] = {'id': 99999}  # Member that does
-    #     # not exist
-    #     activity_list = [null_member_activity]
-    #
-    #     method_name = 'djconnectwise.api.SalesAPIClient.get_activities'
-    #     mock_call, _patch = \
-    #         mocks.create_mock_call(method_name, activity_list)
-    #     synchronizer = sync.ActivitySynchronizer(full=True)
-    #
-    #     created_count, updated_count, deleted_count = \
-    #         synchronizer.sync()
-    #
-    #     # The existing Activity (#47) should be deleted and
-    #     # null_member_activity should not be added to the db
-    #     self.assertEqual(created_count, 0)
-    #     self.assertEqual(updated_count, 0)
-    #     self.assertEqual(deleted_count, 1)
+    def test_sync_null_member_activity(self):
+        null_member_activity = deepcopy(fixtures.API_SALES_ACTIVITY)
+        null_member_activity['id'] = 999
+        null_member_activity['assignTo'] = {'id': 99999}  # Member that does
+        # not exist
+        activity_list = [null_member_activity]
+
+        method_name = 'djconnectwise.api.SalesAPIClient.get_activities'
+        mock_call, _patch = \
+            mocks.create_mock_call(method_name, activity_list)
+        synchronizer = sync.ActivitySynchronizer(full=True)
+
+        created_count, updated_count, deleted_count = \
+            synchronizer.sync()
+
+        # The existing Activity (#47) should be deleted and
+        # null_member_activity should not be added to the db
+        self.assertEqual(created_count, 0)
+        self.assertEqual(updated_count, 0)
+        self.assertEqual(deleted_count, 1)
 
 
 class TestSyncSettings(TestCase):
@@ -1510,3 +1866,66 @@ class TestItemSynchronizer(TestCase, SynchronizerTestMixin):
         self.assertEqual(instance.id, json_data['id'])
         self.assertEqual(instance.name, json_data['name'])
         self.assertEqual(instance.board.name, json_data['board']['name'])
+
+
+class TestWorkTypeSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.WorkTypeSynchronizer
+    model_class = WorkType
+    fixture = fixtures.API_WORK_TYPE_LIST
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_work_types()
+
+    def call_api(self, return_data):
+        return mocks.time_api_get_work_types_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.inactive_flag, json_data['inactiveFlag'])
+
+
+class TestWorkRoleSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.WorkRoleSynchronizer
+    model_class = WorkRole
+    fixture = fixtures.API_WORK_ROLE_LIST
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_work_roles()
+
+    def call_api(self, return_data):
+        return mocks.time_api_get_work_roles_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.inactive_flag, json_data['inactiveFlag'])
+
+
+class TestAgreementSynchronizer(TestCase, SynchronizerTestMixin):
+    synchronizer_class = sync.AgreementSynchronizer
+    model_class = Agreement
+    fixture = fixtures.API_AGREEMENT_LIST
+
+    def setUp(self):
+        super().setUp()
+        fixture_utils.init_agreements()
+        fixture_utils.init_work_roles()
+        fixture_utils.init_work_types()
+
+    def call_api(self, return_data):
+        return mocks.finance_api_get_agreements_call(return_data)
+
+    def _assert_fields(self, instance, json_data):
+        self.assertEqual(instance.id, json_data['id'])
+        self.assertEqual(instance.name, json_data['name'])
+        self.assertEqual(instance.bill_time, json_data['billTime'])
+        self.assertEqual(instance.agreement_type, json_data['type']['name'])
+        self.assertEqual(instance.cancelled_flag, json_data['cancelledFlag'])
+        self.assertEqual(
+            instance.work_role.name, json_data['workRole']['name'])
+        self.assertEqual(
+            instance.work_type.name, json_data['workType']['name'])
+        self.assertEqual(instance.company.name, json_data['company']['name'])
